﻿' This program is able to communicate with the Max-i chip by means of a serial COM port. It is the new version with Local /
'   Global support.
' The input to the transmitter field may be any mix between hexadecimal numbers with an even number of digits and
'   ASCII characters embedded in quotation marks like e.g. "UU" 77 """A" 66 55FF => 55 55 77 22 41 66 55 FF.
'     Note that ASCII " is specified as "" as in Visual Basic. In case of hexadecimal numbers with 4 or more
'       digits, the number is transmitted with the big-endian model, meaning that the most significant part is
'         transmitted first.
' The maximum length of each telegram is 2048 bytes. If longer telegrams are needed, the size of the two arrays
'   RXArray and TXArray must be increased.
'
' This version is able to use the EXAR XR21V141x, XR21B1411 family of USB-serial converters in Wide mode and therefore
'   give a correct telegram separation.
'
' Innovatic, Carsten Kanstrup. March 24th 2016
'                              May 30th 2010
'                              April 14th 2011
'                              August 5th 2011
'                              February 15th 2012
'                              August 14th 2013. Addition of new bitrate row.
'                              August 22nd 2014. New group system.
'                              April 16th 2020. Improved Break generation and shorter Waittime.
'                              September 4th 2021. Added automatic signal generation.
'                              March 31th 2023. Changes for version 13.x
'                              February 9th 2024. Correction of bug with password setting
'                              February 12th 2024. Hardware CTS flow control enabled in PortSelection().
'                              March 15th 2024. Better transmitter sequence to prevent missing first byte and overflow problems with Break.
'
' WARNING! THIS PROGRAM MUST BE COMPILED WITH FRAMEWORK 2.0! NO OTHER TESTED FRAMEWORK WORKS IN A RELIABLE WAY - NOT EVEN 3.5!
'   Go to [Project][MaxiTesterExar Properties...][Compile][Advanced Compile Options][Target Framework] and be sure to
'     select [.NET Framework 2.0]

Imports System
Imports System.IO.Ports
Imports System.Threading
Imports System.Threading.Thread
Imports System.Reflection
Imports System.Runtime.InteropServices
Imports Microsoft.Win32.SafeHandles

' Imports the System, System.IO.Ports and System.Threading.Thread namespaces so that e.g.
'   System.IO.Ports.SerialPort may just be written as SerialPort, and System.EventArgs may just be
'     written as EventArgs.
'       Reflection, Runtime.InteropServices and Microsoft.Win32.SafeHandles are used to communicate with the EXAR driver.

Public Class MaxiTesterExar

    Public Enum GlobalAtt
        Time0 = 0
        Password1 = 1
        Password2Speed = 2
        Password3Speed = 3
        Free4 = 4
        Free5 = 5
        Free6 = 6
        Free7 = 7
    End Enum

    Public Enum LocalAtt
        HourCounter0 = 0
        ProducedValue1 = 1
        ConsumedValue2 = 2
        ErrorWord3 = 3
        ValueID4 = 4
        ProSignature5 = 5
        ConsumerID6 = 6
        ConsumerSignature7 = 7
        Group8 = 8
        Hardware1Setup9 = 9
        Hardware2Setup10 = 10
        Hardware3Setup11 = 11
        IOObjectSetup12 = 12
        SerialNumber13 = 13
        VendorID14 = 14
        HardwareVersion15 = 15
    End Enum

    Dim SpaceCount As Byte = 0
    Dim AttBaseNumber As Integer
    Dim Speed As Integer
    Dim LUT As String = "0123456789ABCDEF"
    Dim RXByte As Byte
    Dim RXByte2 As Byte
    'Dim CrLfInserted As Boolean = False
    Dim RXArray(2052) As Char
    Dim RXPtr As Integer = 0
    Dim OldRXPtr As Integer = 0
    'Dim WriteRX As Boolean = True
    Dim BreakDetected As Boolean = False
    Dim PortHandle As SafeFileHandle
    Dim GlobalAttr As Boolean = False
    Dim LocalAttr As Boolean = False
    Dim Waittime As Byte = 6              ' Default Break time = 6 ms
    Dim ExarWriteBuffer(3) As Byte
    Dim ExarReadBuffer() As Byte = {0, 0, 0, 0}
    Dim inPtr As IntPtr                   ' Pointers to unmanaged memory for the two EXAR buffers
    Dim outPtr As IntPtr
    Dim RedValue As Byte
    Dim GreenValue As Byte
    Dim BlueValue As Byte
    Dim WhiteValue As Byte
    Dim Calculate As UInteger
    Dim Password As String
    Dim Space As String = " "
    Dim Parity As Boolean
    Dim TXFIFOBytes As Byte

    Private Const FILE_DEVICE_UNKNOWN = &H22
    Private Const ExarPortIndex = &H805
    Private Const METHOD_BUFFERED = 0
    Private Const FILE_ANY_ACCESS = 0

    Private Function CTL_CODE(ByVal lngDevFileSys As Long, ByVal lngFunction As Long, _
                          ByVal lngMethod As Long, ByVal lngAccess As Long) As Long
        Return (lngDevFileSys << 16) Or (lngAccess << 14) Or (lngFunction << 2) Or lngMethod
    End Function

    Dim ExarReadCode As Long = CTL_CODE(FILE_DEVICE_UNKNOWN, ExarPortIndex + 16, METHOD_BUFFERED, FILE_ANY_ACCESS)
    Dim ExarWriteCode As Long = CTL_CODE(FILE_DEVICE_UNKNOWN, ExarPortIndex + 17, METHOD_BUFFERED, FILE_ANY_ACCESS)
    Dim ReturnedBytes As Integer

    Dim WithEvents COMPort As New SerialPort ' Make a new System.IO.Ports.SerialPort instance, which is able to send events

    '--------------------------------------------------------------------------------------------------------------------

    Private Sub FillBuffer()
        While (COMPort.BytesToRead > 0) And (RXPtr < 2048) And (BreakDetected = False)
            RXByte = COMPort.ReadByte
            If Not UARTSelected.Checked Then
                RXByte2 = COMPort.ReadByte        ' Get 2nd byte
                If (RXByte = 0) And (RXByte2 = 6) Then
                    BreakDetected = True
                    'OwnTXDetected = True
                Else
                    If RXByte2 <> 0 Then
                        COMPort.DiscardInBuffer()
                        MsgBox("Synchronization lost. Input buffer discarded")
                        RXPtr = 0
                    End If
                End If
            End If
            If Not BreakDetected Then
                OldRXPtr = RXPtr
                RXArray(RXPtr) = LUT(RXByte >> 4)     ' Convert received byte to Hex
                RXPtr = RXPtr + 1
                RXArray(RXPtr) = LUT(RXByte And 15)
                RXPtr = RXPtr + 1
                RXArray(RXPtr) = " "
                RXPtr = RXPtr + 1
                SpaceCount = (SpaceCount + 1) And 31  ' Insert spaces and CRLF for better readability
                If SpaceCount = 0 Then                ' Insert CRLF after 32 numbers
                    RXArray(RXPtr) = Chr(13) ' CR
                    RXPtr = RXPtr + 1
                    RXArray(RXPtr) = Chr(10) ' LF
                    RXPtr = RXPtr + 1
                Else
                    If (SpaceCount And 3) = 0 Then    ' Insert two extra spaces for each 4 numbers
                        RXArray(RXPtr) = " "
                        RXPtr = RXPtr + 1
                        RXArray(RXPtr) = " "
                        RXPtr = RXPtr + 1
                    End If
                End If
            End If
        End While
    End Sub

    ' This subroutine updates the modem control signal lamps, and in UART mode, it triggers a data transfer when a Break
    '   is detected. Note that there is a mutual blocking between all events so that in case of a Break (PinChanged event),
    '     the DataReceived event cannot fire, so the buffer will be emtied.

    Private Sub LampsAndReceive(ByVal sender As Object, ByVal e As SerialPinChangedEventArgs) Handles COMPort.PinChanged
        UpdateLamps()
        If UARTSelected.Checked Then
            If e.EventType = SerialPinChange.Break Then
                FillBuffer()                              ' Get any bytes after the Break event - usually 0-1
                If RXByte = 0 Then RXPtr = OldRXPtr ' Remove 00 Break bytes
                Me.Invoke(New MethodInvoker(AddressOf AppendTextBuffer))
            End If
        End If
    End Sub

    Private Sub UpdateLamps()
        If COMPort.DsrHolding Then
            DSRLamp.BackColor = Color.LightGreen
        Else
            DSRLamp.BackColor = Color.Gray
        End If
        If COMPort.CtsHolding Then
            CTSLamp.BackColor = Color.LightGreen
        Else
            CTSLamp.BackColor = Color.Gray
        End If
    End Sub

    ' This subroutine receives the data by means of the DataReceived event. If EXAR mode is selected, the routine
    '   invokes AppendTextBuffer

    Private Sub ExarReceive(ByVal sender As Object, ByVal e As SerialDataReceivedEventArgs) Handles COMPort.DataReceived
        Do
            Me.Invoke(New MethodInvoker(AddressOf AppendTextBuffer))
        Loop Until COMPort.BytesToRead = 0
    End Sub


    ' Text display routine, which appends the received string to any text in the Received TextBox.

    Private Sub AppendTextBuffer()
        FillBuffer()
        If BreakDetected Then
            RXArray(RXPtr) = Chr(13) ' CR
            RXPtr = RXPtr + 1
            RXArray(RXPtr) = Chr(10) ' LF
            RXPtr = RXPtr + 1
            SpaceCount = 0
            'CrLfInserted = True
            Received.AppendText(New String(RXArray, 0, RXPtr))
            RXPtr = 0
            'WriteRX = True
            BreakDetected = False
        End If
    End Sub

    ' Transmitter subroutine.

    Private Sub Transmit()
        Dim TextString As String
        Dim TXArray(2047) As Byte
        Dim I As Integer
        Dim J As Integer = 0
        Dim Ascii As Boolean = False
        Dim Quote As Boolean = False
        Dim Temp As Boolean
        Dim Second As Boolean = False
        Dim TXByte As Byte = 0
        Dim CharByte As Byte
        If COMPort.IsOpen Then
            TextString = Transmitted.Text
            For I = 0 To TextString.Length - 1
                CharByte = Asc(TextString.Chars(I))
                If CharByte = 34 Then         ' If " Then
                    Temp = Ascii
                    Ascii = Ascii Or Quote
                    Quote = Not (Temp And Quote)
                Else
                    Ascii = Ascii Xor Quote
                    Quote = False
                End If
                If Not Quote Then
                    If Ascii Then
                        TXArray(J) = CharByte
                        J = J + 1
                        If Not UARTSelected.Checked Then
                            TXArray(J) = 0                      ' Insert extra dummy byte
                            J = J + 1
                        End If
                    Else
                        If (CharByte <> 32) And (CharByte <> 10) And (CharByte <> 13) Then ' Skip spaces, LF and CR
                            CharByte = (CharByte - 48) And 31   ' And 31 makes it case insensitive
                            If CharByte > 16 Then
                                CharByte = CharByte - 7
                            End If
                            If Second Then
                                TXArray(J) = TXByte + CharByte
                                Second = False
                                J = J + 1
                                If Not UARTSelected.Checked Then
                                    TXArray(J) = 0               ' Insert extra dummy byte
                                    J = J + 1
                                End If
                            Else
                                TXByte = CharByte << 4
                                Second = True
                            End If
                        End If
                    End If
                End If
            Next
            Try
                COMPort.Write(TXArray, 0, J)
            Catch ex As Exception
                MsgBox(ex.Message & "  Check CTS signal or set Flow Control to None.")
            End Try
            Do While ((ReadReg(&HC81) > 0) Or (COMPort.CtsHolding = False)) ' Wait for TX-FIFO to get empty and CTS = 1
                'Sleep(0)
            Loop
        Else
            MsgBox("COM port is closed. Please select a COM port")
        End If
    End Sub

    ' This subroutine handles a change of COM ports.

    Private Sub PortSelection(ByVal sender As Object, ByVal e As EventArgs) Handles COMPortsBox.SelectedIndexChanged
        RTSLamp.BackColor = Color.Gray
        DTRLamp.BackColor = Color.Gray
        If COMPort.IsOpen Then
            COMPort.RtsEnable = False
            COMPort.DtrEnable = False
            COMPort.Close()
            Application.DoEvents()       ' Empty the message queue before the UI thread goes to sleep to update the modem lamps and the ComboBox.
            Sleep(200)                   ' Wait 0.2 second for port to close as this does not happen immediately.
        End If
        COMPort.PortName = COMPortsBox.Text
        COMPort.BaudRate = 9600          ' Default for Max-i: 9600 bit/s, 8 data bits, no parity, 1 stop bit
        COMPort.Handshake = Handshake.RequestToSend
        ' COMPort.Handshake = Handshake.None
        COMPort.WriteTimeout = 2000      ' Max time to wait for CTS = 2 sec.
        ' COMPort.ReadBufferSize = 16384      Necessary buffer size for 16C950 UART at 921.6 kbit/s
        ' COMPort.ReadTimeout = 5000
        Try
            COMPort.Open()
            If Not UARTSelected.Checked Then
                PortHandle = GetPortHandle(COMPort)        ' Get port handle
                If EXAR21B.Checked Then
                    WriteReg(&HC00, 0)   ' Disable UART before programming
                    WriteReg(&HD02, 1)   ' Set Wide mode for XR21B1411
                    ' WriteReg(&HCC2, 1) ' Set Low Latency mode (default)
                    ' WriteReg(&HC0C, 1) ' Set CTS on GPIO4 and RTS on GPIO5 (default)
                    ' WriteReg(&HC06, 1) ' Enable CTS Hardware Flow Control (default)
                    WriteReg(&HC00, 3)   ' Enable UART after programming
                End If
                If EXAR21V.Checked Then WriteReg(3, 1) ' Set Wide mode for XR21V141x
            End If
        Catch ex As Exception
            MsgBox(ex.Message)
        End Try
        BaudRateBox.Text = COMPort.BaudRate.ToString
        BitsBox.Text = COMPort.DataBits.ToString
        ParityBox.Text = COMPort.Parity.ToString
        FlowControlBox.Text = COMPort.Handshake.ToString
        If COMPort.IsOpen Then
            If COMPort.Handshake <> Handshake.RequestToSend Then COMPort.RtsEnable = True
            UpdateLamps()
            RTSLamp.BackColor = Color.LightGreen
            COMPort.DtrEnable = True
            DTRLamp.BackColor = Color.LightGreen
            COMPort.DiscardInBuffer()
        End If
    End Sub

    ' This subroutine is activated when the form is loaded. It does all the basic initializations.
    ' Note! Baud Rates above 38.4 KBaud requires a special UART like e.g. 16C650, 16C750, 16C850 or 16C950 and
    '   a special driver.
    ' RTS/CTS hardware flow control is necessary for Max-i communication. Note that the standard UART 16C550
    '   has no hardware flow control - except for UART's from Texas Instruments.
    ' The minimum FIFO size is calculated on the basis of a 1 mS response time, which is the absolute minimum
    '   for Windows applications.

    Private Sub MaxiTesterLoad(ByVal sender As Object, ByVal e As EventArgs) Handles MyBase.Load
        For Each COMString As String In My.Computer.Ports.SerialPortNames ' Load all available COM ports.
            COMPortsBox.Items.Add(COMString)
        Next
        COMPortsBox.Sorted = True
        BaudRateBox.Text = "9600"
        BaudRateBox.Items.Add("1200")       ' Single buffer is enough.
        BaudRateBox.Items.Add("2400")       ' Single buffer is enough.
        BaudRateBox.Items.Add("4800")       ' Single buffer is enough.
        BaudRateBox.Items.Add("9600")       ' Single buffer is enough.
        BaudRateBox.Items.Add("19200")      ' Only double buffer required. Can be handled by any UART
        BaudRateBox.Items.Add("38400")
        BaudRateBox.Items.Add("76800")
        BaudRateBox.Items.Add("125000")     ' Min. FIFO size 8 bytes
        BaudRateBox.Items.Add("250000")     ' Min. FIFO size 16 bytes (16C550)
        BaudRateBox.Items.Add("500000")     ' Min. FIFO size 32 bytes (16C650)
        BaudRateBox.Items.Add("1000000")    ' Min. FIFO size 64 bytes (16C750)
        BaudRateBox.Items.Add("2000000")    ' Min. FIFO size 128 bytes (16C850 or 16C950)
        BaudRateBox.Items.Add("4000000")    ' Min. FIFO size 256 bytes
        BitsBox.Items.Add("8")
        ParityBox.Items.Add("None")
        ' ParityBox.Items.Add("Odd")
        ' ParityBox.Items.Add("Even")
        ' ParityBox.Items.Add("Mark")       ' Leaves the parity bit set to 1
        ' ParityBox.Items.Add("Space")      ' Leaves the parity bit set to 0
        FlowControlBox.Items.Add("None")
        FlowControlBox.Items.Add("RequestToSend")
        Password1.Text = "0"
        Password2.Text = "0"
        Password3.Text = "0"
        DataType.Text = "LAMPCTRL"
        DataType.Items.Add("(FIX)")
        DataType.Items.Add("RAW")
        DataType.Items.Add("(FIXBCD)")
        DataType.Items.Add("FLOAT")
        DataType.Items.Add("BCD")
        DataType.Items.Add("ASCII")
        DataType.Items.Add("UNICODE")
        DataType.Items.Add("TIME")
        DataType.Items.Add("LAMPCTRL")
        DataType.Items.Add("PATTERN")
        DataType.Items.Add("UNDEF")
        PNSID.Items.Add("Q1")
        PNSID.Items.Add("Q2")
        PNSID.Items.Add("Q3")
        PNSID.Items.Add("Q4")
        PNSID.Items.Add("Q5")
        PNSID.Items.Add("Q6")
        PNSID.Items.Add("Q7")
        PNSID.Items.Add("Q8")
        PNSID.Items.Add("Q9")
        PNSID.Items.Add("Q10")
        PNSID.Items.Add("Q20")
        PNSID.Items.Add("Q30")
        PNSID.Items.Add("Q40")
        PNSID.Items.Add("Q50")
        PNSID.Items.Add("Q60")
        PNSID.Items.Add("Q70")
        PNSID.Items.Add("Q80")
        PNSID.Items.Add("Q90")
        PNSID.Items.Add("O0")
        PNSID.Items.Add("O1")
        PNSID.Items.Add("O2")
        PNSID.Items.Add("O3")
        PNSID.Items.Add("O4")
        PNSID.Items.Add("O5")
        PNSID.Items.Add("O6")
        PNSID.Items.Add("O7")
        PNSID.Items.Add("O8")
        PNSID.Items.Add("O9")
        PNSID.Items.Add("O255")
        PNSID.Text = "O255"
        AttBase.Items.Add("63")
        AttBase.Items.Add("0")
        AttBase.Items.Add("1")
        AttBase.Items.Add("2")
        AttBase.Items.Add("3")
        AttBase.Items.Add("4")
        AttBase.Items.Add("5")
        AttBase.Items.Add("6")
        AttBase.Items.Add("7")
        AttBase.Items.Add("8")
        AttBase.Items.Add("9")
        AttBase.Text = "63"
        RTSLamp.BackColor = Color.Gray
        DTRLamp.BackColor = Color.Gray
        CTSLamp.BackColor = Color.Gray
        DSRLamp.BackColor = Color.Gray
        MinimumVal.Items.Add("0")
        MinimumVal.Items.Add("15")
        MinimumVal.Items.Add("31")
        MinimumVal.Items.Add("47")
        MinimumVal.Items.Add("63")
        MinimumVal.Items.Add("79")
        MinimumVal.Items.Add("95")
        MinimumVal.Items.Add("111")
        MinimumVal.Items.Add("127")
        MinimumVal.Items.Add("143")
        MinimumVal.Items.Add("159")
        MinimumVal.Items.Add("175")
        MinimumVal.Items.Add("191")
        MinimumVal.Items.Add("207")
        MinimumVal.Items.Add("223")
        MinimumVal.Items.Add("239")
        MaximumVal.Items.Add("15")
        MaximumVal.Items.Add("31")
        MaximumVal.Items.Add("47")
        MaximumVal.Items.Add("63")
        MaximumVal.Items.Add("79")
        MaximumVal.Items.Add("95")
        MaximumVal.Items.Add("111")
        MaximumVal.Items.Add("127")
        MaximumVal.Items.Add("143")
        MaximumVal.Items.Add("159")
        MaximumVal.Items.Add("175")
        MaximumVal.Items.Add("191")
        MaximumVal.Items.Add("207")
        MaximumVal.Items.Add("223")
        MaximumVal.Items.Add("239")
        MaximumVal.Items.Add("255")
        Received.AppendText(vbCrLf & "           Select a UART mode and then select a COM Port and flow control" & vbCrLf)
        inPtr = Marshal.AllocHGlobal(4)        ' Allocate unmanaged memory to hold 4-byte arrays for EXAR communication
        outPtr = Marshal.AllocHGlobal(4)
        AutoSignature.Checked = True
        TelegramNumber.Text = "1"
        PasswordGen(Password1.Text)
        Signature.Text = Password
    End Sub

    ' This subroutine is used to close the COM port. Because the receiver uses Invoke instead of BeginInvoke,
    '   this subroutine is usually called on a separate (new) thread to prevent a close-down deadlock.

    Private Sub ClosePort()
        If COMPort.IsOpen Then COMPort.Close()
    End Sub

    ' This subroutine is activated when the form is closed. It closes the COM port. Without such a close command,
    '   the garbage collector may close the COM port while it is still in use!

    Private Sub MaxiTesterClosing(ByVal sender As Object, ByVal e As ComponentModel.CancelEventArgs) Handles MyBase.Closing
        If MessageBox.Show("Do you want to close the window", "", MessageBoxButtons.YesNo) = Windows.Forms.DialogResult.No Then
            e.Cancel = True
        Else
            Marshal.FreeHGlobal(inPtr)               ' Free allocated unmanaged memory
            Marshal.FreeHGlobal(outPtr)
            Dim t As New Thread(AddressOf ClosePort) ' Close COM port on a new thread when the form is terminated with [X]
            t.Start()
        End If
    End Sub

    ' This subroutine clears the Received TextBox for received bytes.

    Private Sub ClearReceivedText(ByVal sender As Object, ByVal e As EventArgs) Handles ClearButton.Click
        Received.Text = ""
        SpaceCount = 0
    End Sub

    ' This subroutines sends a Break condition.

    Private Sub SendBreak()
        If COMPort.IsOpen Then
            If UARTSelected.Checked Then
                COMPort.BreakState = True
                Application.DoEvents()               ' Empty message queue so that BreakState is executed and the following delay is not visible
                Sleep(Waittime)
                COMPort.BreakState = False
            End If
            If EXAR21B.Checked Then
                Do While COMPort.CtsHolding = False  ' Wait for CTS = 1
                    Sleep(0)
                Loop
                WriteReg(&HC0A, Waittime)            ' Generate a Break
                Sleep(Waittime)                      ' Wait for Break to be active or else the first byte of the message may be transfered just before the Break,
                '                                      which resets the Max-i UART, and the first byte of a message could therefore be missing.
            End If
            If EXAR21V.Checked Then
                Do While COMPort.CtsHolding = False  ' Wait for CTS = 1
                    Sleep(0)
                Loop
                WriteReg(&H14, Waittime)             ' XR21V141x does not allow read of number of bytes in FIFO and does not automatically time the Break (time = don't care)
                Sleep(Waittime)
                WriteReg(&H14, 0)
            End If
        Else
            MsgBox("No COM Port Selected")
        End If
    End Sub

    Private Sub LoadData()
        Dim Index As UInt32
        Dim SumValue As Integer
        Dim Chr6Val As Integer = 0
        Dim Chr0, Chr1, Chr2, Chr3, Chr4, Chr5, Chr6 As Char
        If Bit0.Checked Then Chr6Val = Chr6Val Or 8
        If Bit1.Checked Then Chr6Val = Chr6Val Or 4
        If Bit2.Checked Then Chr6Val = Chr6Val Or 2
        If Bit3.Checked Then Chr6Val = Chr6Val Or 1
        Chr6 = LUT(Chr6Val)
        If GlobalAttSelect.Checked Then
            Index = [Enum].Parse(GetType(GlobalAtt), Attribute.Text)
            Select Case Index
                Case 0
                    Dim DayTime As UInt16
                    Dim Nibble As UInt16
                    Dim DT As DateTime = DateTime.Parse(Data.Text)
                    DayTime = (DT.Day << 11) + (DT.Hour << 6) + DT.Minute
                    Nibble = (DayTime >> 12) + 48
                    If Nibble > 57 Then
                        Nibble = Nibble + 7
                    End If
                    Chr0 = Chr(Nibble)
                    Nibble = ((DayTime >> 8) And 15) + 48
                    If Nibble > 57 Then
                        Nibble = Nibble + 7
                    End If
                    Chr1 = Chr(Nibble)
                    Nibble = ((DayTime >> 4) And 15) + 48
                    If Nibble > 57 Then
                        Nibble = Nibble + 7
                    End If
                    Chr2 = Chr(Nibble)
                    Nibble = (DayTime And 15) + 48
                    If Nibble > 57 Then
                        Nibble = Nibble + 7
                    End If
                    Chr3 = Chr(Nibble)
                    DayTime = DT.Second << 2
                    Nibble = (DayTime >> 4) + 48
                    If Nibble > 57 Then
                        Nibble = Nibble + 7
                    End If
                    Chr4 = Chr(Nibble)
                    Nibble = (DayTime And 15) + 48
                    If Nibble > 57 Then
                        Nibble = Nibble + 7
                    End If
                    Chr5 = Chr(Nibble)
                    Transmitted.Text = "00 00 " & Chr0 & Chr1 & " " & Chr2 & Chr3 & " " & Chr4 & Chr5 _
                    & " 00 " & Chr6 & "F 00 80"
                Case 1
                    PasswordGen(Password1.Text)
                    Transmitted.Text = "00 0" & CStr(Index) & " " & Data.Text & " " & Chr6 & Password
                Case 2
                    PasswordGen(Password2.Text)
                    Transmitted.Text = "00 0" & CStr(Index) & " " & Data.Text & " " & Chr6 & Password
                Case 3
                    PasswordGen(Password3.Text)
                    Transmitted.Text = "00 0" & CStr(Index) & " " & Data.Text & " " & Chr6 & Password
                Case 4
                    Transmitted.Text = "00 0" & CStr(Index) & " " & Data.Text & " " & Chr6 & "F 80 ED"
                Case 5
                    Transmitted.Text = "00 0" & CStr(Index) & " " & Data.Text & " " & Chr6 & "F 80 A2"
                Case 6
                    Transmitted.Text = "00 0" & CStr(Index) & " " & Data.Text & " " & Chr6 & "F 80 83"
                Case 7
                    Transmitted.Text = "00 0" & CStr(Index) & " " & Data.Text & " " & Chr6 & "F 80 CC"
                Case Else
                    Transmitted.Text = "00 0" & CStr(Index) & " " & Data.Text & " " & Chr6 & "F FF FF"
            End Select
        End If
        If BroadcastBoolean.Checked Then
            SumValue = PNSConvert()
            If Local.Checked Then
                SumValue = SumValue Or &H8000
            End If
            Transmitted.Text = LUT((SumValue >> 12) And &HF) & LUT((SumValue >> 8) And &HF) & " " & LUT((SumValue >> 4) And &HF) & LUT(SumValue And &HF) _
                & " " & Chr6 & Signature.Text
        End If
        If Broadcast20.Checked Or Broadcast36.Checked Then
            SumValue = PNSConvert()
            If Local.Checked Then
                SumValue = SumValue Or &H8000
            End If
            Transmitted.Text = LUT((SumValue >> 12) And &HF) & LUT((SumValue >> 8) And &HF) & " " _
               & LUT((SumValue >> 4) And &HF) & LUT(SumValue And &HF) & " " & Data.Text & " " & Chr6 & Signature.Text
        End If
        If PollSelect.Checked Then
            SumValue = PNSConvert()
            If Local.Checked Then
                SumValue = SumValue Or &H8000
            End If
            Transmitted.Text = LUT((SumValue >> 12) And &HF) & LUT((SumValue >> 8) And &HF) & " " & LUT((SumValue >> 4) And &HF) & LUT(SumValue And &HF) & " "
        End If
        If LocalAttPollSelect.Checked Then
            Index = [Enum].Parse(GetType(LocalAtt), Attribute.Text)
            SumValue = (((Val(AttBase.Text) << 4) + Index) Or &HFC00) << 15
            SumValue = SumValue + PNSConvert()
            Transmitted.Text = "F" & LUT((SumValue >> 24) And &HF) & " " & LUT((SumValue >> 20) And &HF) & LUT((SumValue >> 16) And &HF) & " " _
                & LUT((SumValue >> 12) And &HF) & LUT((SumValue >> 8) And &HF) & " " & LUT((SumValue >> 4) And &HF) & LUT(SumValue And &HF) & " "
        End If
        If LocalAttWriteSelect.Checked Then
            Index = [Enum].Parse(GetType(LocalAtt), Attribute.Text)
            SumValue = (((Val(AttBase.Text) << 4) + Index) Or &HFC00) << 15
            SumValue = SumValue + PNSConvert()
            Transmitted.Text = "7" & LUT((SumValue >> 24) And &HF) & " " & LUT((SumValue >> 20) And &HF) & LUT((SumValue >> 16) And &HF) & " " _
                & LUT((SumValue >> 12) And &HF) & LUT((SumValue >> 8) And &HF) & " " & LUT((SumValue >> 4) And &HF) & LUT(SumValue And &HF) & " " _
                & Data.Text & " " & Chr6 & Signature.Text
        End If
    End Sub

    Private Sub Break(ByVal sender As Object, ByVal e As EventArgs) Handles BreakButton.Click
        SendBreak()
    End Sub

    Private Sub BreakLoad(ByVal sender As Object, ByVal e As EventArgs) Handles BreakLoadButton.Click
        SendBreak()
        LoadData()
    End Sub

    Private Sub Send(ByVal sender As Object, ByVal e As EventArgs) Handles SendButton.Click
        Transmit()
    End Sub

    Private Sub BreakSend(ByVal sender As Object, ByVal e As EventArgs) Handles BreakSendButton.Click
        SendBreak()
        Transmit()
    End Sub

    Private Sub BreakLoadSend(ByVal sender As Object, ByVal e As EventArgs) Handles BreakLoadSendButton.Click
        SendBreak()
        LoadData()
        Transmit()
    End Sub

    ' This subroutine handles a change in Baud Rate.

    Private Sub BaudRateSelection(ByVal sender As Object, ByVal e As EventArgs) Handles BaudRateBox.SelectedIndexChanged
        Speed = CInt(BaudRateBox.Text)
        COMPort.BaudRate = Speed
        Waittime = (11000 \ COMPort.BaudRate)     ' Ensure minimum 11 bit Break delay
        If Waittime = 0 Then
            Waittime = 1                          ' 1 - 2 ms
        End If
    End Sub

    ' This subroutine handles a change in number of data bits

    Private Sub DataBitsSelection(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles BitsBox.SelectedIndexChanged
        COMPort.DataBits = CInt(BitsBox.Text)
    End Sub

    ' This subroutine handles a change in the parity.

    Private Sub ParitySelection(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ParityBox.SelectedIndexChanged
        COMPort.Parity = CType([Enum].Parse(GetType(Parity), ParityBox.Text), Parity)
    End Sub

    ' This subroutine handles a change in the flow control (handshake).

    Private Sub SoftwareFlowControlSelection(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles FlowControlBox.SelectedIndexChanged
        COMPort.Handshake = CType([Enum].Parse(GetType(Handshake), FlowControlBox.Text), Handshake)
    End Sub

    ' This subroutine saves the content of the Received TextBox

    Private Sub SaveText(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles SaveAs.Click
        Dim SaveFileDialog1 As New SaveFileDialog()
        SaveFileDialog1.Filter = "Text Files (*.txt)|*.txt"
        SaveFileDialog1.Title = "Save Received As"
        If SaveFileDialog1.ShowDialog() = System.Windows.Forms.DialogResult.OK _
            And SaveFileDialog1.FileName.Length > 0 Then
            My.Computer.FileSystem.WriteAllText(SaveFileDialog1.FileName, Received.Text, False) ' Overwrite file
        End If
    End Sub

    Private Sub GlobalAttNoChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Attribute.TextChanged
        SetSignature()
    End Sub

    Private Sub SetSignature()
        Dim FirstChar As Char
        Parity = False
        If GlobalAttSelect.Checked Then
            Select Case [Enum].Parse(GetType(GlobalAtt), Attribute.Text)
                Case 0
                    Data.Text = DateTime.Now.ToString
                    Signature.Text = "F 00 80"
                Case 1
                    PasswordGen(Password1.Text)
                    If Parity Then
                        Data.Text = "7" & Password & " FF"
                    Else
                        Data.Text = "F" & Password & " FF"
                    End If
                    Signature.Text = Password
                Case 2
                    Dim SpeedCode As Byte = 15
                    Dim SpeedFactor As Integer = Speed \ 1200
                    While SpeedFactor > 0
                        SpeedFactor = SpeedFactor >> 1
                        SpeedCode = SpeedCode - 1
                    End While
                    Dim SpeedParityEven As String = "F77F7FF77FF7F77F"
                    Dim SpeedParityOdd As String = "7FF7F77FF77F7FF7"
                    PasswordGen(Password2.Text)
                    If Parity Then
                        FirstChar = SpeedParityOdd(SpeedCode)
                    Else
                        FirstChar = SpeedParityEven(SpeedCode)
                    End If
                    SpeedCode = SpeedCode + 48
                    If SpeedCode > 57 Then
                        SpeedCode = SpeedCode + 7
                    End If
                    Data.Text = FirstChar & Password & " F" & Chr(SpeedCode)
                    Signature.Text = Password
                Case 3
                    PasswordGen(Password3.Text)
                    If Parity Then
                        Data.Text = "F" & Password & " FB"
                    Else
                        Data.Text = "7" & Password & " FB"
                    End If
                    Signature.Text = Password
                Case 4
                    Data.Text = "FF FF FF FF"
                    Signature.Text = "F 80 ED"
                Case 5
                    Data.Text = "FF FF FF FF"
                    Signature.Text = "F 80 A2"
                Case 6
                    Data.Text = "FF FF FF FF"
                    Signature.Text = "F 80 83"
                Case 7
                    Data.Text = "FF FF FF FF"
                    Signature.Text = "F 80 CC"
                Case Else
                    Data.Text = "FF FF FF FF"
                    Signature.Text = "F FF FF"
            End Select
        Else
            Select Case [Enum].Parse(GetType(LocalAtt), Attribute.Text)
                Case 0 To 3
                    PasswordGen(Password1.Text)
                Case 4 To 9
                    PasswordGen(Password2.Text)
                Case 10 To 15
                    PasswordGen(Password3.Text)
            End Select
            Signature.Text = Password
        End If
    End Sub

    Private Sub GlobalAttSelected(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles GlobalAttSelect.CheckedChanged
        Transmitted.Text = ""
        If GlobalAttSelect.Checked Then
            Code.Text = "0000 0000 0000 0"
            Code.Visible = True
            AttBase.Visible = False
            If GlobalAttr = False Then
                Attribute.Items.Clear()
                For Each s As String In [Enum].GetNames(GetType(GlobalAtt))
                    Attribute.Items.Add(s)
                Next
                Attribute.Text = "Time0"
                GlobalAttr = True
                LocalAttr = False
            End If
            Attribute.Visible = True
            Local.Visible = False
            PNSID.Visible = False
            Data.Text = DateTime.Now.ToString
            Data.Visible = True
            Bit0.Visible = True
            Bit1.Visible = True
            Bit2.Visible = True
            Bit3.Visible = True
            Signature.Text = "F 00 80"
            Signature.Visible = True
            AutoSignature.Visible = False
            IntensityHide()
        End If
    End Sub

    Private Sub BroadcastBooleanSelected(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles BroadcastBoolean.CheckedChanged
        Transmitted.Text = ""
        If BroadcastBoolean.Checked Then
            Code.Visible = False
            AttBase.Visible = False
            Attribute.Visible = False
            Local.Visible = True
            PNSID.Visible = True
            Data.Visible = False
            Bit0.Visible = True
            Bit1.Visible = True
            Bit2.Visible = True
            Bit3.Visible = True
            '           Signature.Text = "F 00 FF"  ' TIME
            Signature.Visible = True
            AutoSignature.Visible = True
            IntensityHide()
            CalculateSignature()
        End If
    End Sub

    Private Sub Broadcast20Selected(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Broadcast20.CheckedChanged
        Transmitted.Text = ""
        If Broadcast20.Checked Then
            Code.Visible = False
            AttBase.Visible = False
            Attribute.Visible = False
            Local.Visible = True
            PNSID.Visible = True
            Data.Text = "00 00"
            Data.Visible = True
            Bit0.Visible = True
            Bit1.Visible = True
            Bit2.Visible = True
            Bit3.Visible = True
            '            Signature.Text = "F 80 FF"  ' PATTERN
            Signature.Visible = True
            AutoSignature.Visible = True
            RedBar.Visible = False
            GreenBar.Visible = False
            BlueBar.Visible = True
            WhiteBar.Visible = True
            MinimumVal.Visible = True
            MaximumVal.Visible = True
            Intensity.Visible = True
            SelectedColor.Visible = True
            ColorChanged()
            CalculateSignature()
        End If
    End Sub

    Private Sub Broadcast36Selected(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Broadcast36.CheckedChanged
        Transmitted.Text = ""
        If Broadcast36.Checked Then
            Code.Visible = False
            AttBase.Visible = False
            Attribute.Visible = False
            Local.Visible = True
            PNSID.Visible = True
            Data.Text = "00 00 00 00"
            Data.Visible = True
            Bit0.Visible = True
            Bit1.Visible = True
            Bit2.Visible = True
            Bit3.Visible = True
            Signature.Visible = True
            AutoSignature.Visible = True
            RedBar.Visible = True
            GreenBar.Visible = True
            BlueBar.Visible = True
            WhiteBar.Visible = True
            MinimumVal.Visible = True
            MaximumVal.Visible = True
            Intensity.Visible = True
            SelectedColor.Visible = True
            ColorChanged()
            CalculateSignature()
        End If
    End Sub


    Private Sub PollSelected(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles PollSelect.CheckedChanged
        Transmitted.Text = ""
        If PollSelect.Checked Then
            Code.Visible = False
            AttBase.Visible = False
            Attribute.Visible = False
            Local.Visible = True
            PNSID.Visible = True
            Data.Visible = False
            Bit0.Visible = False
            Bit1.Visible = False
            Bit2.Visible = False
            Bit3.Visible = False
            Signature.Visible = False
            AutoSignature.Visible = False
            IntensityHide()
        End If
    End Sub

    Private Sub LocalAttPollSelected(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles LocalAttPollSelect.CheckedChanged
        Transmitted.Text = ""
        If LocalAttPollSelect.Checked Then
            Code.Text = "1111 111"
            Code.Visible = True
            AttBase.Visible = True
            If LocalAttr = False Then
                Attribute.Items.Clear()
                For Each s As String In [Enum].GetNames(GetType(LocalAtt))
                    Attribute.Items.Add(s)
                Next
                Attribute.Text = "HourCounter0"
                LocalAttr = True
                GlobalAttr = False
            End If
            Attribute.Visible = True
            Local.Visible = False
            PNSID.Visible = True
            Data.Visible = False
            Bit0.Visible = False
            Bit1.Visible = False
            Bit2.Visible = False
            Bit3.Visible = False
            Signature.Visible = False
            AutoSignature.Visible = False
            IntensityHide()
        End If
    End Sub

    Private Sub LocalAttWriteSelected(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles LocalAttWriteSelect.CheckedChanged
        Transmitted.Text = ""
        If LocalAttWriteSelect.Checked Then
            Code.Text = "0111 111"
            Code.Visible = True
            AttBase.Visible = True
            If LocalAttr = False Then
                Attribute.Items.Clear()
                For Each s As String In [Enum].GetNames(GetType(LocalAtt))
                    Attribute.Items.Add(s)
                Next
                Attribute.Text = "HourCounter0"
                LocalAttr = True
                GlobalAttr = False
            End If
            Attribute.Visible = True
            Local.Visible = False
            PNSID.Visible = True
            Data.Text = "FF FF FF FF"
            Data.Visible = True
            Bit0.Visible = True
            Bit1.Visible = True
            Bit2.Visible = True
            Bit3.Visible = True
            SetSignature()
            Signature.Visible = True
            AutoSignature.Visible = False
            IntensityHide()
        End If
    End Sub

    Private Sub AsciiHexSelected(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles AsciiHexSelect.CheckedChanged
        Transmitted.Text = ""
        If AsciiHexSelect.Checked Then
            Code.Visible = False
            AttBase.Visible = False
            Attribute.Visible = False
            Local.Visible = False
            PNSID.Visible = False
            Data.Visible = False
            Bit0.Visible = False
            Bit1.Visible = False
            Bit2.Visible = False
            Bit3.Visible = False
            Signature.Visible = False
            AutoSignature.Visible = False
            IntensityHide()
        End If
    End Sub

    Private Function PNSConvert() As Integer
        Dim PNSIDUpper As String = PNSID.Text.ToUpper
        If PNSIDUpper = "" Then PNSIDUpper = "O255"
        If PNSIDUpper.StartsWith("Q") And Char.IsDigit(PNSIDUpper, 1) Then ' Short Identifier Qxxxx
            PNSIDUpper = PNSIDUpper.Remove(0, 1)
            Return (((Val(PNSIDUpper) And &HFFF) << 3) Or 7)               ' Works only on short messages
        End If
        If PNSIDUpper.StartsWith("O") And Char.IsDigit(PNSIDUpper, 1) Then ' Short Identifier Oxxx = group
            PNSIDUpper = PNSIDUpper.Remove(0, 1)
            Return (Val(PNSIDUpper) And &HFF) + &H7D00
        End If
    End Function

    '-------------------------------------------------------------------------------------------------------------------
    ' Functions for communication with the EXAR driver

    Private Function GetPortHandle(ByVal Port As SerialPort) As SafeFileHandle
        Dim BaseStream As Object = Port.BaseStream
        Dim BaseStreamType As Type = BaseStream.GetType
        Return BaseStreamType.GetField("_handle", BindingFlags.NonPublic Or BindingFlags.Instance).GetValue(BaseStream)
    End Function

    Declare Function DeviceIoControl Lib "kernel32.dll" (ByVal hDevice As SafeFileHandle, _
                                                         ByVal IoControlCode As Integer, _
                                                         ByVal InBuffer As IntPtr, _
                                                         ByVal InBufferSize As Integer, _
                                                         ByVal OutBuffer As IntPtr, _
                                                         ByVal OutBufferSize As Integer, _
                                                         ByRef BytesReturned As Integer, _
                                                         ByVal Overlapped As IntPtr) As Integer

    Private Sub WriteReg(ByVal Reg As UInt16, ByVal Data As UInt16)
        ExarWriteBuffer(0) = Reg And &HFF  ' Little endian model
        ExarWriteBuffer(1) = Reg >> 8
        ExarWriteBuffer(2) = Data And &HFF ' Little endian model
        ExarWriteBuffer(3) = Data >> 8
        Marshal.Copy(ExarWriteBuffer, 0, inPtr, 4)
        Dim Status As Boolean = DeviceIoControl(PortHandle, ExarWriteCode, inPtr, 4, -1, 0, ReturnedBytes, 0)
        If Not Status Then
            MsgBox("Error while writing Exar USB-UART register")
        End If
    End Sub

    Private Function ReadReg(ByVal Reg As UInt16) As UInt16
        ExarWriteBuffer(0) = Reg And &HFF  ' Little endian model
        ExarWriteBuffer(1) = Reg >> 8
        ExarWriteBuffer(2) = 0
        ExarWriteBuffer(3) = 0
        Marshal.Copy(ExarWriteBuffer, 0, inPtr, 4)
        Dim Status As Boolean = DeviceIoControl(PortHandle, ExarReadCode, inPtr, 4, outPtr, 4, ReturnedBytes, 0)
        If Not Status Then
            MsgBox("Error while reading Exar USB-UART register")
        End If
        Marshal.Copy(outPtr, ExarReadBuffer, 0, 2)
        Return (ExarReadBuffer(1) << 8) Or ExarReadBuffer(0)
    End Function

    Private Sub IntensitySend()
        Calculate = ((RedBar.Value + 1) * Intensity.Value) >> 8
        RedValue = CByte(Calculate)
        Calculate = ((GreenBar.Value + 1) * Intensity.Value) >> 8
        GreenValue = CByte(Calculate)
        Calculate = ((BlueBar.Value + 1) * Intensity.Value) >> 8
        BlueValue = CByte(Calculate)
        Calculate = ((WhiteBar.Value + 1) * Intensity.Value) >> 8
        WhiteValue = CByte(Calculate)
        Data.Text = ""
        If Broadcast36.Checked Then
            Data.Text = LUT(RedValue >> 4) & LUT(RedValue And 15) & " " & LUT(GreenValue >> 4) & LUT(GreenValue And 15) & " "
        End If
        Data.Text = Data.Text & LUT(BlueValue >> 4) & LUT(BlueValue And 15) & " " & LUT(WhiteValue >> 4) & LUT(WhiteValue And 15)
        SendBreak()
        LoadData()
        Transmit()
    End Sub

    Private Sub ColorChanged()
        SelectedColor.Text = ""
        If Broadcast36.Checked Then
            SelectedColor.Text = LUT(RedBar.Value >> 4) & LUT(RedBar.Value And 15) & " " & LUT(GreenBar.Value >> 4) & LUT(GreenBar.Value And 15) & " "
        End If
        SelectedColor.Text = SelectedColor.Text & LUT(BlueBar.Value >> 4) & LUT(BlueBar.Value And 15) & " " & LUT(WhiteBar.Value >> 4) & LUT(WhiteBar.Value And 15)
    End Sub

    Private Sub Intensity_Scroll(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Intensity.Scroll
        IntensitySend()
    End Sub

    Private Sub MinimumVal_Changed(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MinimumVal.SelectedIndexChanged
        Intensity.Minimum = CInt(MinimumVal.Text)
        If Intensity.Value < Intensity.Minimum Then Intensity.Value = Intensity.Minimum
    End Sub

    Private Sub MaximumVal_Changed(sender As Object, e As EventArgs) Handles MaximumVal.SelectedIndexChanged
        Intensity.Maximum = CInt(MaximumVal.Text)
        If Intensity.Value > Intensity.Maximum Then Intensity.Value = Intensity.Maximum
    End Sub

    Private Sub IntensityHide()
        MinimumVal.Visible = False
        MaximumVal.Visible = False
        Intensity.Visible = False
        RedBar.Visible = False
        GreenBar.Visible = False
        BlueBar.Visible = False
        WhiteBar.Visible = False
        SelectedColor.Visible = False
    End Sub

    Private Sub RedBar_Scroll(sender As Object, e As EventArgs) Handles RedBar.Scroll
        ColorChanged()
        IntensitySend()
    End Sub

    Private Sub GreenBar_Scroll(sender As Object, e As EventArgs) Handles GreenBar.Scroll
        ColorChanged()
        IntensitySend()
    End Sub

    Private Sub BlueBar_Scroll(sender As Object, e As EventArgs) Handles BlueBar.Scroll
        ColorChanged()
        IntensitySend()
    End Sub

    Private Sub WhiteBar_Scroll(sender As Object, e As EventArgs) Handles WhiteBar.Scroll
        ColorChanged()
        IntensitySend()
    End Sub

    Private Sub CalculateSignature()
        Dim SumValue As Integer
        Dim CheckBits As Byte
        Dim SigValue As Integer
        Dim CheckWord1 As Integer
        Dim CheckWord2 As Integer
        Dim CheckWord3 As Integer
        Dim CheckWord4 As Integer
        Dim CheckWord5 As Integer
        Dim CheckWord6 As Integer
        Dim CheckWord7 As Integer
        SumValue = PNSConvert() << 16 Or &HFFFF     ' Works only on short identifiers
        CheckWord1 = SumValue And &H7FFFFFFF
        CheckWord2 = SumValue And &H69969669
        CheckWord3 = SumValue And &HFFFF
        CheckWord4 = SumValue And &HFF00FF
        CheckWord5 = SumValue And &HF0F0F0F
        CheckWord6 = SumValue And &H33333333
        CheckWord7 = SumValue And &H55555555
        CheckBits = 0
        For I = 0 To 30
            If (CheckWord1 And 1) = 1 Then CheckBits = CheckBits Xor 64
            CheckWord1 = CheckWord1 >> 1
            If (CheckWord2 And 1) = 1 Then CheckBits = CheckBits Xor 32
            CheckWord2 = CheckWord2 >> 1
            If (CheckWord3 And 1) = 1 Then CheckBits = CheckBits Xor 16
            CheckWord3 = CheckWord3 >> 1
            If (CheckWord4 And 1) = 1 Then CheckBits = CheckBits Xor 8
            CheckWord4 = CheckWord4 >> 1
            If (CheckWord5 And 1) = 1 Then CheckBits = CheckBits Xor 4
            CheckWord5 = CheckWord5 >> 1
            If (CheckWord6 And 1) = 1 Then CheckBits = CheckBits Xor 2
            CheckWord6 = CheckWord6 >> 1
            If (CheckWord7 And 1) = 1 Then CheckBits = CheckBits Xor 1
            CheckWord7 = CheckWord7 >> 1
        Next
        If DataType.Text = "(FIX)" Then SigValue = 0
        If DataType.Text = "RAW" Then SigValue = &H58000
        If DataType.Text = "(FIXBCD)" Then SigValue = &H80000
        If DataType.Text = "FLOAT" Then SigValue = &HE0000
        If DataType.Text = "BCD" Then SigValue = &HE4000
        If DataType.Text = "ASCII" Then SigValue = &HE8000
        If DataType.Text = "UNICODE" Then SigValue = &HEC000
        If DataType.Text = "TIME" Then SigValue = &HF0000
        If DataType.Text = "LAMPCTRL" Then SigValue = &HF4000
        If DataType.Text = "PATTERN" Then SigValue = &HF8000
        If DataType.Text = "UNDEF" Then SigValue = &HFC000
        If TelegramNumber.Text = "" Then TelegramNumber.Text = "1"
        SigValue = SigValue Or ((CInt(TelegramNumber.Text) And &H7F) << 7)
        If AutoSignature.Checked Then
            SigValue = SigValue Or CheckBits
        Else
            SigValue = SigValue Or &H7F
        End If
        Signature.Text = LUT((SigValue >> 16) And 15) & " " & LUT((SigValue >> 12) And 15) & LUT((SigValue >> 8) And 15) & " " & LUT((SigValue >> 4) And 15) & LUT(SigValue And 15)
    End Sub

    Private Sub PNSID_SelectedIndexChanged(sender As Object, e As EventArgs) Handles PNSID.SelectedIndexChanged
        CalculateSignature()
    End Sub

    Private Sub DataType_SelectedIndexChanged(sender As Object, e As EventArgs) Handles DataType.SelectedIndexChanged
        If DataType.Text = "(FIX)" Then
            TelegramNumber.Text = "0"
        Else
            TelegramNumber.Text = "1"
        End If
        CalculateSignature()
    End Sub

    Private Sub AutoSignature_CheckedChanged(sender As Object, e As EventArgs) Handles AutoSignature.CheckedChanged
        CalculateSignature()
    End Sub

    Private Sub TelegramNumber_TextChanged(sender As Object, e As EventArgs) Handles TelegramNumber.TextChanged
        CalculateSignature()
    End Sub

    ' This subroutine generates the global parameter "Password" and modifies the global parameter "Parity" according to the parity of the password.

    Private Sub PasswordGen(ByVal PasswordText As String)
        Dim PasswordHex As Int32 = (CUInt(PasswordText) Xor &HFFFFFFFF) And &HFFFFF
        Dim PasswordInv As String = Hex(PasswordHex)
        Dim PasswordInvA As String = PasswordInv.Insert(1, Space)
        Password = PasswordInvA.Insert(4, Space)
        For I = 0 To 19
            If (PasswordHex And 1) = 1 Then Parity = Not Parity
            PasswordHex = PasswordHex >> 1
        Next
    End Sub

End Class
